#include "stdafx.h"
#include "BmsConverter.hpp"
#include "O2JamNoteValues.hpp"
#include "SoundFormat.hpp"
#include "AlphaDecimal.hpp"
#include "NoteChannels.hpp"
#include "SoundFlags.hpp"
#include "Application.hpp"
#include "Utility.hpp"

namespace Lmt {
namespace O2Bms {
	//08/4/20 1.05----->
	const wchar_t *BmsConverter::m_difficaltyO2[] = {
		L" (Easy)",
		L" (Normal)",
		L" (Hard)",
	};
	const wchar_t *BmsConverter::m_difficaltyIIDX[] = {
		L" -NORMAL-",
		L" -HYPER-",
		L" -ANOTHER-",
	};
	const wchar_t *BmsConverter::m_difficaltyNone[] = {
		L"",
		L"",
		L"",
	};
	//<-----------------

	bool BmsConverter::Convert( BmsConverter::FromO2Jam &desc, const O2JamMusic &m ) const {
		O2JamNote note;
		O2JamMusicDesc md = m.GetDescription();
		int levels[3] = { md.GetEasyLevel(), md.GetNormalLevel(), md.GetHardLevel() };
		//2007/5/20 1.02----->
		int level;
		//<-------------------
		//08/4/20 1.05----->
//		wstring difficulties[3] = { L"Easy", L"Normal", L"Hard" };
		const wchar_t **pdiffTypes[3] = { m_difficaltyO2, m_difficaltyIIDX, m_difficaltyNone };
		wstring difficalty;
		wstring difficaltyFname;
		//<-----------------
		wstring inputFiles[3] = { desc.inputEasyNote, desc.inputNormalNote, desc.inputHardNote };
		int realLevel;
		wstring fileName;
		wstring path;
		//2007/2/24 1.01----->
		int keyCount;
		//<-------------------

		desc.notFoundSoundCount = 0;
		desc.outOfRankSoundCount = 0;
		
		for (int i = 0; i < 3; i++) {
			if (inputFiles[i].length() == 0) {
				continue;
			}

			if (BmsConverter::O2BuildNote( inputFiles[i], note ) == false) {
				return false;
			}

			//08/4/20 1.05----->
			difficalty.assign( pdiffTypes[desc.diffType][i] );
			//<-----------------
			
			//2007/5/20 1.02----->
			level = min( levels[i], desc.maxLevelO2 );
//			realLevel = (int)((double)levels[i] * (desc.maxLevel / (double)BmsConverter::MaxLevel));
			realLevel = (int)((double)levels[i] * ((double)desc.maxLevel / (double)desc.maxLevelO2));
//			realLevel = min( realLevel, BmsConverter::MaxLevel );
			realLevel = min( realLevel, desc.maxLevel );
			//<-------------------
			realLevel = max( realLevel, BmsConverter::MinLevel );
			//08/4/20 1.05----->
//			fileName = BmsConverter::O2BuildFileName( md.GetTitle(), difficulties[i] );
			difficaltyFname = difficalty;
			if (difficaltyFname == L"") {
				difficaltyFname = pdiffTypes[O2JamDifficulty][i];
			}
			fileName = BmsConverter::O2BuildFileName( md.GetTitle(), difficaltyFname );
			//<-----------------
			path = Utility::CombinePath( desc.outputDir, fileName );
			
			{
				wstringstream script;
				//2007/2/24 1.01----->
				wstringstream scriptHeader;
				//<-------------------
				wfstream fsbms( Utility::WStringToString( path ).c_str(), ios::out );
				
				if (fsbms.is_open() == false) {
					return false;
				}
				
				//2007/2/24 1.01----->
//				BmsConverter::O2AddBmsHeader( script, md, realLevel, desc.judgeRank, difficulties[i] );
				//<-------------------
				BmsConverter::O2AddBmsSounds( script, desc );
				BmsConverter::O2AddBmsTempos( script, note );
				//2007/2/24 1.01----->
				BmsConverter::O2AddBmsNote( script, note, desc, keyCount );
				//08/4/20 1.05----->
//				BmsConverter::O2AddBmsHeader( scriptHeader, md, realLevel, desc.judgeRank, O2CalculateTotalGroove( keyCount ), difficulties[i] );
				BmsConverter::O2AddBmsHeader( scriptHeader, md, realLevel, desc.judgeRank, O2CalculateTotalGroove( keyCount ), difficalty );
				//<-----------------
				//<-------------------
				
				//2007/2/24 1.01----->
//				fsbms << script.str();
				fsbms << scriptHeader.str() << script.str();
				//<-------------------
				if (fsbms.fail()) {
					return false;
				}
				
				fsbms.close();
			}
		}
		return true;
	}
	
	wstring BmsConverter::O2BuildFileName( const wstring &title, const wstring &difficulty ) {
		wstringstream ss;
		
		ss << title;
		//08/4/20 1.05----->
//		ss << L" (" << difficulty << L").bms";
		ss << difficulty << L".bms";
		//<-----------------
		
		return ss.str();
	}
	
	bool BmsConverter::O2BuildNote( const wstring &path, O2JamNote &note ) {
		fstream fs( Utility::WStringToString( path ).c_str(), ios::in | ios::binary );
		O2JamNoteHeader hdr;
		NoteValue nval;
		
		if (fs.is_open() == false) {
			return false;
		}
		
		note.Clear();
		while (fs.eof() == false) {
			fs.read( (char*)&hdr, sizeof(hdr) );
			if (fs.gcount() != sizeof(hdr)) {
				break;
			}
			
			{
				O2JamNoteValues values( hdr.channel );
				
				for (int i = 0; i < hdr.count; i++) {
					fs.read( (char*)&nval.value, sizeof(nval.value) );
					if (fs.gcount() != sizeof(nval.value)) {
						break;
					}
					values.Add( nval );
				}
				
				note.Add( hdr.measure, values );
			}
		}
		note.BestCondition();
		
		fs.close();
		return true;
	}
	
	//2007/2/24 1.01----->
//	void BmsConverter::O2AddBmsHeader( wstringstream &ss, const O2JamMusicDesc &desc, int level, int judgeRank, const wstring &difficulty ) {
	void BmsConverter::O2AddBmsHeader( wstringstream &ss, const O2JamMusicDesc &desc, int level, int judgeRank, int totalGroove, const wstring &difficulty ) {
	//<-------------------
		ss << L"; O2Jam - BMS" << endl;
		ss << L";" << endl;
		//2007/5/20 1.02----->
//		ss << L"; Created by O2BMS ver1.00" << endl;
		//2007/5/23 1.03----->
//		ss << L"; Created by O2BMS ver1.02" << endl;
		//08/4/20 1.05----->
//		ss << L"; Created by O2BMS ver1.03" << endl;
		ss << L"; Created by O2BMS ver1.05" << endl;
		//<-----------------
		//<-------------------
		//<-------------------
		ss << L";" << endl;
		ss << endl;
		
		ss << L";Header------------------------------" << endl;
		ss << L"#PLAYER 1" << endl;
		ss << L"#GENRE " << desc.GetGenre() << endl;
		//08/4/20 1.05----->
//		ss << L"#TITLE " << desc.GetTitle() << L" (" << difficulty << L")" << endl;
		ss << L"#TITLE " << desc.GetTitle() << difficulty << endl;
		//<-----------------
		ss << L"#ARTIST " << desc.GetComposer() << endl;
		ss << L"#MAKER " << desc.GetScoreMaker() << endl;
		ss << L"#BPM " << desc.GetInitialBPM() << endl;
		ss << L"#PLAYLEVEL " << level << endl;
		ss << L"#RANK " << judgeRank << endl;
		//2007/2/24 1.01----->
//		ss << L"#TOTAL " << Application::TotalGroove << endl;
		ss << L"#TOTAL " << totalGroove << endl;
		//<-------------------
		ss << L"#LNTYPE 1" << endl;
		//2007/5/20 1.02----->
//		ss << L"#LNOBJ ZZ" << endl;
		//<-------------------
		ss << L"#STAGEFILE songimg.bmp" << endl;
		ss << endl;
	}
	
	void BmsConverter::O2AddBmsSounds( wstringstream &ss, const FromO2Jam &o2 ) {
		wstring identifier;
		wstring formatExt = SoundFormat( o2.format ).GetExtension();
		int maxSoundCount = min( o2.maxSoundCount, o2.psoundRanking->GetCount() );
		
		ss << L";Sound materials---------------------" << endl;
		for (int i = 0; i < maxSoundCount; i++) {
			identifier = AlphaDecimal( i + 1 ).ToString();
			ss << L"#WAV" << identifier << L" " << identifier << formatExt << endl;
		}

		identifier = AlphaDecimal( o2.dummySoundIndex ).ToString();
		ss << L"#WAV" << identifier << L" " << identifier << SoundFormat( SoundFormat::Wave ).GetExtension() << endl;
		ss << endl;
	}
	
	void BmsConverter::O2AddBmsTempos( wstringstream &ss, const O2JamNote &note ) {
		int count = note.GetTempoCount();
		if (count == 0) {
			return;
		}
		
		vector<float> tempos( count );
		vector<int> ids( count );
		note.GetTempoData( &(tempos.at( 0 )), &(ids.at( 0 )) );
		
		ss << L";BPM materials-----------------------" << endl;
		for (int i = 0; i < count; i++) {
			ss << L"#BPM";
			ss << AlphaDecimal( ids[i] ).ToString() << L" ";
			ss << tempos[i];
			ss << endl;
		}
		ss << endl;
	}
	
	//2007/2/24 1.01----->
//	void BmsConverter::O2AddBmsNote( wstringstream &ss, const O2JamNote &note, FromO2Jam &o2 ) {
	void BmsConverter::O2AddBmsNote( wstringstream &ss, const O2JamNote &note, FromO2Jam &o2, int &keyCount ) {
	//<-------------------
		int start = note.GetMinMeasure();
		int finish = note.GetMaxMeasure();
		//2007/2/24 1.01----->
		int count = 0;
		int countTemp;
		//<-------------------
		//08/4/20 1.05----->
		bool lnHolds[7] = { false, false, false, false, false, false, false };
		//<-----------------
		
		ss << L";Notes-------------------------------" << endl;
		for (int i = start; i <= finish; i++) {
			ss << L";[" << i << "]" << endl;
			BmsConverter::O2AddBmsNoteBpm( ss, note, o2, i );
			//2007/2/24 1.01----->
//			BmsConverter::O2AddBmsNotePlayable( ss, note, o2, i );
			//08/4/20 1.05----->
//			BmsConverter::O2AddBmsNotePlayable( ss, note, o2, i, countTemp );
			BmsConverter::O2AddBmsNotePlayable( ss, note, o2, i, lnHolds, countTemp );
			//<-----------------
			//<-------------------
			BmsConverter::O2AddBmsNoteBackground( ss, note, o2, i );
			ss << endl;
			count += countTemp;
		}
		ss << endl;

		//2007/2/24 1.01----->
		keyCount = count;
		//<-------------------
	}
	
	void BmsConverter::O2AddBmsNoteBpm( wstringstream &ss, const O2JamNote &note, FromO2Jam &o2, int measure ) {
		const O2JamNoteMeasure *pm = note.GetTempoMeasure( measure );
		if (pm == NULL || pm->GetCount() == 0) {
			return;
		}
		
		const O2JamNoteValues *pv = pm->GetValues( 0 );
		if (pv->GetChannel() != ChangeBpm) {
			return;
		}
		
		ss << BmsConverter::O2MakeNoteDef( measure, 8 );
		
		for (int i = 0; i < pv->GetResolution(); i++) {
			int identifier = note.GetTempoIdentifier( pv->GetValue( i ).tempo );
			ss << ((identifier == -1) ? L"00" : AlphaDecimal( identifier ).ToString());
		}
		
		ss << endl;
	}
	
	//2007/2/24 1.01----->
//	void BmsConverter::O2AddBmsNotePlayable( wstringstream &ss, const O2JamNote &note, FromO2Jam &o2, int measure ) {
	//08/4/20 1.05----->
//	void BmsConverter::O2AddBmsNotePlayable( wstringstream &ss, const O2JamNote &note, FromO2Jam &o2, int measure, int &keyCount ) {
	void BmsConverter::O2AddBmsNotePlayable( wstringstream &ss, const O2JamNote &note, FromO2Jam &o2, int measure, bool *lnHolds, int &keyCount ) {
	//<-----------------
	//<-------------------
		static int bmsKeyChannels[7] = { 11, 12, 13, 14, 15, 18, 19 };
		static int bmsLnChannels[7] = { 51, 52, 53, 54, 55, 58, 59 };
		const O2JamNoteMeasure *pm = note.GetPlayableMeasure( measure );
		//2007/2/24 1.01----->
		int count = 0;
		//<-------------------
		//08/4/20 1.05----->
		int trackIdx;
		//<-----------------

		if (pm == NULL) {
			//2007/2/24 1.01----->
			keyCount = 0;
			//<-------------------
			return;
		}
		
		for (int i = 0; i < pm->GetCount(); i++) {
			const O2JamNoteValues *pv = pm->GetValues( i );
			int channel = pv->GetChannel();
			
			if (channel < PlayableKey1 || channel > PlayableKey7) {
				continue;
			}
			//08/4/20 1.05----->
			else {
				trackIdx = channel - PlayableKey1;
			}
			//<-----------------
			
			wstringstream sskey;
			wstringstream ssln;
			bool hasKey = false;
			bool hasLn = false;
			NoteValue nval;
			int idOrg;
			int idFind;
			int rank;

			//08/4/20 1.05----->
//			sskey << BmsConverter::O2MakeNoteDef( measure, bmsKeyChannels[channel - PlayableKey1] );
//			ssln << BmsConverter::O2MakeNoteDef( measure, bmsLnChannels[channel - PlayableKey1] );
			sskey << BmsConverter::O2MakeNoteDef( measure, bmsKeyChannels[trackIdx] );
			ssln << BmsConverter::O2MakeNoteDef( measure, bmsLnChannels[trackIdx] );
			//<-----------------

			for (int j = 0; j < pv->GetResolution(); j++) {
				nval = pv->GetValue( j );
				idOrg = nval.sound.identifier;
				idFind = idOrg + o2.bgmCount;

				//2007/2/24 1.01----->
//				if ((nval.sound.flags & LongNoteDef) == 0) {
				if ((nval.sound.flags & LongNoteDef) == 0 || o2.nothingLongNote == true) {
				//<-------------------
					//2007/2/24 1.01----->
//					if (idOrg != 0) {
					if (idOrg != 0 && (nval.sound.flags & (LongNoteEnd)) == 0) {
					//<-------------------
						hasKey = true;
						ssln << L"00";
						if ((rank = o2.psoundRanking->GetRank( idFind )) == SoundRanking::NotFound) {
							o2.notFoundSoundCount++;
							sskey << AlphaDecimal( o2.dummySoundIndex ).ToString();
						}
						else if (rank > o2.maxSoundCount) {
							o2.outOfRankSoundCount++;
							sskey << AlphaDecimal( o2.dummySoundIndex ).ToString();
						}
						else {
							sskey << AlphaDecimal( rank ).ToString();
						}
						//2007/2/24 1.01----->
						count++;
						//<-------------------
					}
					else {
						ssln << L"00";
						sskey << L"00";
					}
				}
				else if (nval.sound.flags & LongNoteEnd) {
					//08/4/20 1.05----->
//					hasLn = true;
//					sskey << L"00";
//					ssln << L"ZZ";
					if (lnHolds[trackIdx] == true) {
						lnHolds[trackIdx] = false;
						hasLn = true;
						sskey << L"00";
						ssln << L"ZZ";
					}
					else {
						sskey << L"00";
						ssln << L"00";
					}
					//<-----------------
				}
				else {
					//08/4/20 1.05----->
//					if (idOrg != 0) {
					if ((idOrg != 0) && (lnHolds[trackIdx] == false)) {
					//<-----------------
						hasLn = true;
						//08/4/20 1.05----->
						lnHolds[trackIdx] = true;
						//<-----------------
						sskey << L"00";
						if ((rank = o2.psoundRanking->GetRank( idFind )) == SoundRanking::NotFound) {
							o2.notFoundSoundCount++;
							ssln << AlphaDecimal( o2.dummySoundIndex ).ToString();
						}
						else if (rank > o2.maxSoundCount) {
							o2.outOfRankSoundCount++;
							ssln << AlphaDecimal( o2.dummySoundIndex ).ToString();
						}
						else {
							ssln << AlphaDecimal( rank ).ToString();
						}
						//2007/2/24 1.01----->
						count++;
						//<-------------------
					}
					else {
						sskey << L"00";
						ssln << L"00";
					}
				}
			}
			
			if (hasKey == true) {
				ss << sskey.str() << endl;
			}
			if (hasLn == true) {
				ss << ssln.str() << endl;
			}

			//2007/2/24 1.01----->
			keyCount = count;
			//<-------------------
		}
	}
	
	void BmsConverter::O2AddBmsNoteBackground( wstringstream &ss, const O2JamNote &note, FromO2Jam &o2, int measure ) {
		const O2JamNoteMeasure *pm;
		const O2JamNoteValues *pv;
		wstringstream ssLine;
		NoteValue nval;
		int identifier;
		int rank;
		bool used;
		
		if ((pm = note.GetBackgroundMeasure( measure )) == NULL) {
			return;
		}

		for (int i = 0; i < pm->GetCount(); i++) {
			pv = pm->GetValues( i );

			ssLine.str( L"" );
			ssLine << BmsConverter::O2MakeNoteDef( measure, 1 );

			used = false;
			for (int j = 0; j < pv->GetResolution(); j++) {
				nval = pv->GetValue( j );

				if (nval.sound.identifier == 0) {
					ssLine << L"00";
				}
				else {
					used = true;

					if (nval.sound.flags & BackgroundMusic) {
						identifier = nval.sound.identifier;
					}
					else {
						identifier = nval.sound.identifier + o2.bgmCount;
					}

					if ((rank = o2.psoundRanking->GetRank( identifier )) == SoundRanking::NotFound) {
						o2.notFoundSoundCount++;
						ssLine << L"00";
					}
					else if (rank > o2.maxSoundCount) {
						o2.outOfRankSoundCount++;
						ssLine << AlphaDecimal( o2.dummySoundIndex ).ToString();
					}
					else {
						ssLine << AlphaDecimal( rank ).ToString();
					}
				}
			}

			if (used == true) {
				ss << ssLine.str() << endl;
			}
		}
	}
	
	wstring BmsConverter::O2MakeNoteDef( int measure, int channel ) {
		wstringstream ss;
		ss << L"#";
		ss << dec;
		ss.width( 3 );
		ss.fill( L'0' );
		ss << measure;
		ss << dec;
		ss.width( 2 );
		ss.fill( L'0' );
		ss << channel;
		ss << dec;
		ss << L":";
		return ss.str();
	}

	//2007/2/24 1.01----->
	int BmsConverter::O2CalculateTotalGroove( int keyCount ) {
		if (keyCount < 120) {
			return 150 + (int)(((double)keyCount / 120.0) * 150.0);
		}
		else {
			return 300 + (int)(((double)keyCount - 120.0) / 3.6);
		}
	}
	//<-------------------
}
}
