#include "stdafx.h"
#include "MusicPackageStream.hpp"
#include "Utility.hpp"

namespace Lmt {
namespace O2Bms {
	//08/4/20 1.05----->
//	char MusicPackageStream::m_xor[] = { 'n', 'a', 'm', 'i' };
	const char MusicPackageStream::m_nami[] = { 'n', 'a', 'm', 'i' };
	const char MusicPackageStream::m_noenc[] = { '\0', '\0', '\0', '\0' };
	//<-----------------

	//08/4/20 1.05----->
//	MusicPackageStream::MusicPackageStream( fstream *pfs ) {
	MusicPackageStream::MusicPackageStream( fstream *pfs, bool decrypt ) {
	//<-----------------
		ItemHeader header;
		int basePosition = 0;
		const int maxLength = 1024 * 1024 * 40;  //OGG̋eTCY

		memset( &m_header, 0, sizeof(header) );
		m_basePosition = 0;
		m_nowPosition = 0;
		m_rank = 0;
		m_pfs = NULL;

		basePosition = pfs->tellg();
		basePosition += sizeof(header);

		pfs->read( (char*)&header, sizeof(header) );
		if (pfs->gcount() != sizeof(header)) {
			return;
		}

		if (header.Identifier < 0) {
			return;
		}
		if (header.Length <= 0 || header.Length > maxLength) {
			return;
		}
		if (header.PcmLength <= 0) {
			return;
		}
		header.Name[sizeof(header.Name) - 1] = '\0';

		pfs->seekg( header.Length, ios::cur );

		m_header = header;
		m_basePosition = basePosition;
		m_nowPosition = 0;
		m_pfs = pfs;
		//08/4/20 1.05----->
		m_pxor = (decrypt == true) ? m_nami : m_noenc;
		//<-----------------
	}

	bool MusicPackageStream::Available() const {
		return m_pfs != NULL;
	}

	score_t MusicPackageStream::GetSoundScore() const {
		if (this->IsBackgroundMusic() == true) {
			return 0x8000000000000000ULL + m_header.Identifier;
		}
		else {
			return m_header.PcmLength;
		}
	}

	wstring MusicPackageStream::GetName() const {
		return Utility::StringToWString( m_header.Name );
	}

	int MusicPackageStream::GetLength() const {
		return m_header.Length;
	}

	int MusicPackageStream::GetIdentifier() const {
		//OJÑm[g`ID+1`ŋL^Ă邽+1B
		return m_header.Identifier + 1;
	}

	int MusicPackageStream::GetRank() const {
		return m_rank;
	}

	int MusicPackageStream::GetPcmLength() const {
		return m_header.PcmLength;
	}

	bool MusicPackageStream::IsBackgroundMusic() const {
		return (m_header.Flags & IsBgmFlag) != 0;
	}

	int MusicPackageStream::Read( void *pbuffer, int length ) {
		if (this->Available() == false) {
			return -1;
		}
		if (pbuffer == NULL) {
			return -2;
		}
		if (length < 0) {
			return -3;
		}

		length = min( length, (m_header.Length - m_nowPosition) );

		if (length <= 0) {
			return 0;
		}

		m_pfs->seekg( m_basePosition + m_nowPosition, ios::beg );
		m_pfs->read( reinterpret_cast<char*>( pbuffer ), length );

		int readLength = m_pfs->gcount();
		char *buffer = reinterpret_cast<char*>( pbuffer );

		for (int i = 0; i < readLength; i++) {
			//08/4/20 1.05----->
//			buffer[i] = buffer[i] ^ m_xor[(m_nowPosition + i) & 0x03];
			buffer[i] = buffer[i] ^ m_pxor[(m_nowPosition + i) & 0x03];
			//<-----------------
		}

		m_nowPosition += readLength;
		return readLength;
	}

	int MusicPackageStream::Seek( int offset, SeekOrigin origin ) {
		int newPosition;

		switch (origin) {
			case SeekFromBegin:
				newPosition = max( 0, min( m_header.Length, offset ) );
				break;
			case SeekFromEnd:
				newPosition = max( 0, min( m_header.Length, m_header.Length + offset ) );
				break;
			case SeekFromCurrent:
				newPosition = max( 0, min( m_header.Length, m_nowPosition + offset ) );
				break;
			default:
				return -1;
		}

		m_nowPosition = newPosition;
		return 0;
	}

	int MusicPackageStream::Tell() const {
		return m_nowPosition;
	}

	bool MusicPackageStream::operator >( const MusicPackageStream &stream ) const {
		return this->GetSoundScore() > stream.GetSoundScore();
	}
}
}

