#include "stdafx.h"
#include "O2JamNote.hpp"
#include "NoteChannels.hpp"

namespace Lmt {
namespace O2Bms {
	O2JamNote::O2JamNote() {
		this->Clear();
	}

	const O2JamNoteMeasure * O2JamNote::GetTempoMeasure( int measure ) const {
		MeasureMap::const_iterator it = m_tempo.find( measure );
		return it == m_tempo.end() ? NULL : &it->second;
	}

	const O2JamNoteMeasure * O2JamNote::GetPlayableMeasure( int measure ) const {
		MeasureMap::const_iterator it = m_playable.find( measure );
		return it == m_playable.end() ? NULL : &it->second;
	}

	const O2JamNoteMeasure * O2JamNote::GetBackgroundMeasure( int measure ) const {
		MeasureMap::const_iterator it = m_background.find( measure );
		return it == m_background.end() ? NULL : &it->second;
	}

	int O2JamNote::GetMinMeasure() const {
		return m_minMeasure;
	}

	int O2JamNote::GetMaxMeasure() const {
		return m_maxMeasure;
	}
	
	int O2JamNote::GetTempoCount() const {
		return static_cast<int>( m_tempos.size() );
	}
	
	int O2JamNote::GetTempoIdentifier( float tempo ) const {
		TempoMap::const_iterator it = m_tempos.find( tempo );
		return it == m_tempos.end() ? -1 : it->second;
	}
	
	void O2JamNote::GetTempoData( float *ptempos, int *pids ) const {
		int i;
		TempoMap::const_iterator it;
		for (i = 0, it = m_tempos.begin(); it != m_tempos.end(); it++, i++) {
			ptempos[i] = it->first;
			pids[i] = it->second;
		}
	}
	
	int O2JamNote::GetSoundCount() const {
		return static_cast<int>( m_sounds.size() );
	}
	
	int O2JamNote::GetSoundIdentifier( int sound ) const {
		SoundMap::const_iterator it = m_sounds.find( sound );
		return it == m_sounds.end() ? -1 : it->second;
	}
	
	void O2JamNote::GetSoundData( int *psounds, int *pids ) const {
		int i;
		SoundMap::const_iterator it;
		for (i = 0, it = m_sounds.begin(); it != m_sounds.end(); it++, i++) {
			psounds[i] = it->first;
			pids[i] = it->second;
		}
	}

	bool O2JamNote::Add( int measure, const O2JamNoteValues &values ) {
		MeasureMap *pmap;
		MeasureMap::iterator it;
		NoteValue nval;
		int channel = values.GetChannel();

		switch (channel) {
			case ChangeBpm:
				pmap = &m_tempo;
				break;
			case PlayableKey1:
			case PlayableKey2:
			case PlayableKey3:
			case PlayableKey4:
			case PlayableKey5:
			case PlayableKey6:
			case PlayableKey7:
				pmap = &m_playable;
				break;
			default:
				pmap = &m_background;
				break;
		}

		it = pmap->find( measure );
		if (it == pmap->end()) {
			it = pmap->insert( make_pair( measure, O2JamNoteMeasure() ) ).first;
			if (it == pmap->end()) {
				return false;
			}
		}
		it->second.Add( values );
		
		for (int i = 0; i < values.GetResolution(); i++) {
			nval = values.GetValue( i );
			if (channel == ChangeBpm) {
				if (nval.tempo != 0.0f) {
					m_tempos.insert( make_pair( nval.tempo, static_cast<int>( m_tempos.size() + 1 ) ) );
				}
			}
			else {
				if (nval.sound.identifier != 0) {
					m_sounds.insert( make_pair( nval.sound.identifier, nval.sound.identifier ) );
				}
			}
		}

		if (m_minMeasure > measure) {
			m_minMeasure = measure;
		}
		if (m_maxMeasure < measure) {
			m_maxMeasure = measure;
		}
		return true;
	}
	
	void O2JamNote::Clear() {
		m_minMeasure = 0x7fffffff;
		m_maxMeasure = 0;
		m_tempo.clear();
		m_playable.clear();
		m_background.clear();
	}

	void O2JamNote::BestCondition() {
		MeasureMap::iterator it;
		int i;

//		for (it = m_tempo.begin(); it != m_tempo.end(); it++) {
//			for (i = 0; i < it->second.GetCount(); i++) {
//				it->second.GetValues( i )->Expand( 2 );
//			}
//		}
		for (it = m_playable.begin(); it != m_playable.end(); it++) {
			for (i = 0; i < it->second.GetCount(); i++) {
				it->second.GetValues( i )->Optimize();
			}
		}
		for (it = m_background.begin(); it != m_background.end(); it++) {
			for (i = 0; i < it->second.GetCount(); i++) {
				it->second.GetValues( i )->Optimize();
			}
		}
	}
}
}
