#include "stdafx.h"
#include "O2JamNoteValues.hpp"

namespace Lmt {
namespace O2Bms {
	O2JamNoteValues::O2JamNoteValues( int channel ) {
		this->Initialize( channel, NULL, 0 );
	}

	O2JamNoteValues::O2JamNoteValues( int channel, const NoteValue *pvalue, int count ) {
		this->Initialize( channel, pvalue, count );
	}

	int O2JamNoteValues::GetChannel() const {
		return m_channel;
	}

	int O2JamNoteValues::GetResolution() const {
		return static_cast<int>( m_values.size() );
	}

	NoteValue O2JamNoteValues::GetValue( int index ) const {
		return m_values[index];
	}

	void O2JamNoteValues::Add( NoteValue value ) {
		m_values.push_back( value );
	}

	void O2JamNoteValues::Add( const NoteValue *pvalue, int count ) {
		for (int i = 0; i < count; i++) {
			m_values.push_back( pvalue[i] );
		}
	}

	void O2JamNoteValues::Clear() {
		m_values.clear();
	}

	bool O2JamNoteValues::Optimize() {
		if (m_values.size() % 2 != 0) {
			return false;
		}

		size_t i;
		bool isChanged = false;
		ValueArray optimized;
		ValueArray optimizing;

		optimizing.swap( m_values );
		while (optimizing.size() % 2 == 0) {
			for (i = 1; i < optimizing.size(); i+= 2) {
				if (optimizing[i].sound.identifier != 0) {
					goto EndOfOptimize;
				}
			}

			optimized.clear();
			optimized.reserve( optimizing.size() / 2 );
			for (i = 0; i < optimizing.size(); i += 2) {
				optimized.push_back( optimizing[i] );
			}

			optimizing.swap( optimized );
			isChanged = true;
		}

	EndOfOptimize:
		m_values.swap( optimizing );
		return isChanged;
	}

	bool O2JamNoteValues::Expand( int ratio ) {
		static int expandables[] = { 1, 2, 4, 8, 16, 32, 64, 96, 0 };
		int count = this->GetResolution();
		int i, j;
		ValueArray after;

		for (i = 0; expandables[i] != 0; i++) {
			if (count == expandables[i]) {
				break;
			}
		}
		if (expandables[i] == 0) {
			return false;
		}
		if (expandables[i] * ratio > 192) {
			return false;
		}

		after.reserve( expandables[i] * ratio );
		for (i = 0; i < count; i++) {
			after.push_back( m_values[i] );
			for (j = 0; j < ratio - 1; j++) {
				after.push_back( NoteValue( 0 ) );
			}
		}

		m_values.swap( after );
		return true;
	}

	void O2JamNoteValues::Initialize( int channel, const NoteValue *pvalue, int count ) {
		m_channel = channel;

		if (pvalue != NULL && count > 0) {
			m_values.reserve( count );
			for (int i = 0; i < count; i++) {
				m_values.push_back( pvalue[i] );
			}
		}
	}
}
}
